/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

/*- Header ------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <CSet.h>
#include <CmMessage.h>

#include <Db.h>

#include <DbCVS.h>
#include <DbLock.h>

/*---------------------------------------------------------------------*/
#define DBOBJECTSROOT "DBOBJECTSROOT"
#define DBCVSROOT     "DBCVSROOT"
#define CVSROOT       "CVSROOT"
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
DbObject DbLoadObjectCopyFromText (char* text);
int DbLoadObjectCopiesFromText (char* text, CLList list);
DbObject DbLoadObjectFromText     (char* text);
DbObject DbParserCreateObject     (char* name);
char*    BuildFileName            (char* object_name, char* path_name);
char*    BuildTmpFileName         (char* object_name, char* path_name, 
                                   char* suffix);
int      DbSendBreak              (char* name, int transaction);
char*    BuildFullName (char* object_name, char* path_name);
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
static int   CVSConfigured (void);

static char* DbObjectsRoot (void);
static char* ObjectFileName (char* name);

static char* CVSObjectFileName        (char* name);
static char* CVSRoot                  (void);
static char* CVSObjectModuleName      (void);
static char* CVSObjectPathAndFileName (char* path, char* name);

static char* CVSObjectPath (char* path);
static char* CVSObjectFile (char* path, char* name);
static void  CVSAddPath    (char* path);

static void     CVSAdd               (char* path, char* name);
static void     CVSRelease           ();

static DbObject CVSLoadObject        (char* command);
static DbObject CVSLoadObjectByDate  (char* path, char* name, char* date);
static DbObject CVSLoadObjectByTag   (char* path, char* name, char* tag);

static char*    CVSLoadObjects       (char* path_reg, char* name_reg);
static char*    CVSLoadObjectsByDate (char* path_reg, char* name_reg, char* date);
static char*    CVSLoadObjectsByTag  (char* path_reg, char* name_reg, char* tag);

static void     CVSTagObject         (char* path, char* name, char* tag);
static void     CVSTagInPath         (char* path, char* tag);
static void     CVSTagAll            (char* tag);
static void     CVSDeleteTag         (char* path, char* name, char* tag);

static void     CVSRemovePath        (char* path);

static void     CVSCreateBranch      (char* tag, char* branch);
static void     CVSObjectBranch      (char* name, char* branch);
static void     CVSCommitObjectInBranch (DbObject object, char* branch);

static char* DbExecuteCommand (char* command);

/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
static int UseCVS = 0;
/*---------------------------------------------------------------------*/

/*----- Handlers -----------------------------------------------------*/
static CmMessageStatus SetVersioningHandler   (CmMessage message, 
                                               char* sender, char* server);
static CmMessageStatus SetNoVersioningHandler (CmMessage message, 
                                               char* sender, char* server);
static CmMessageStatus GetVersioningHandler   (CmMessage message, 
                                               char* sender, char* server);


static CmMessageStatus TagObjectHandler        (CmMessage message, 
                                                char* sender, char* server);
static CmMessageStatus TagObjectsInPathHandler (CmMessage message, 
                                                char* sender, char* server);
static CmMessageStatus TagDatabaseHandler      (CmMessage message, 
                                                char* sender, char* server);
static CmMessageStatus DeleteObjectTagHandler  (CmMessage message, 
                                                char* sender, char* server);
static CmMessageStatus LoadObjectFromBranchHandler (CmMessage message, 
                                                    char* sender, 
                                                    char* server);

static CmMessageStatus LoadObjectByTagHandler   (CmMessage message, 
                                                 char* sender, char* server);
static CmMessageStatus LoadObjectByDateHandler  (CmMessage message, 
                                                 char* sender, char* server);
static CmMessageStatus LoadObjectsByTagHandler  (CmMessage message, 
                                                 char* sender, char* server);
static CmMessageStatus LoadObjectsByDateHandler (CmMessage message, 
                                                 char* sender, char* server);

static CmMessageStatus CreateBranchHandler (CmMessage message, 
                                            char* sender, char* server);
static CmMessageStatus ObjectBranchHandler (CmMessage message, 
                                            char* sender, char* server);

static CmMessageStatus CommitObjectInBranchHandler (CmMessage message,
                                                    char* sender, char* server);

static CmMessageStatus GetVersionHistoryHandler (CmMessage message, 
                                                 char* sender, char* server);
static CmMessageStatus GetObjectHistoryHandler  (CmMessage message, 
                                                 char* sender, char* server);


static CmMessageStatus CommitObjectHandler (CmMessage message,
                                            char* sender, char* server);
static CmMessageStatus ReleaseHandler      (CmMessage message,
                                            char* sender, char* server);
static CmMessageStatus CVSDefaultHandler   (CmMessage message,
                                            char* sender, char* server);

/*=====================================================================*/


/*---------------------------------------------------------------------*/
void DbInstallVersioningHandler (void)
/*---------------------------------------------------------------------*/
{
  CmMessageInstallHandler (SetVersioningHandler,   "DbSetVersioning");
  CmMessageInstallHandler (SetNoVersioningHandler, "DbSetNoVersioning");
  CmMessageInstallHandler (GetVersioningHandler,   "DbGetVersioning");
}  


/*---------------------------------------------------------------------*/
void DbServerInstallVersioning (void)
/*---------------------------------------------------------------------*/
{
  CVSConfigured ();
  
  if (DbUseCVS ())
    {
      CmMessageInstallHandler (TagObjectHandler,         "DbTagObject");
      CmMessageInstallHandler (TagObjectsInPathHandler,  "DbTagObjectsInPath");
      CmMessageInstallHandler (TagDatabaseHandler,       "DbTagDatabase");
      CmMessageInstallHandler (DeleteObjectTagHandler,   "DbDeleteObjectTag");
      
      CmMessageInstallHandler (LoadObjectByTagHandler,   "DbLoadObjByTag");
      CmMessageInstallHandler (LoadObjectByDateHandler,  "DbLoadObjByDate");
      CmMessageInstallHandler (LoadObjectFromBranchHandler, 
                               "DbLoadObjFromBranch");
      CmMessageInstallHandler (LoadObjectsByTagHandler,  "DbLoadObjsByTag");
      CmMessageInstallHandler (LoadObjectsByDateHandler, "DbLoadObjsByDate");
      
      CmMessageInstallHandler (CreateBranchHandler,      "DbCreateBranch");
      CmMessageInstallHandler (ObjectBranchHandler,      "DbObjectBranch");  /* ?? */
      
      CmMessageInstallHandler (CommitObjectInBranchHandler,
                                                         "DbCiObjInBranch");  /* ?? */

      CmMessageInstallHandler (GetVersionHistoryHandler,
                                                         "DbGetVersionHistory");
      CmMessageInstallHandler (GetObjectHistoryHandler,  "DbGetObjectHistory"); /* ?? */
    }
  else
    {
      CmMessageInstallHandler (CVSDefaultHandler,  "DbTagObject");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbTagObjectsInPath");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbTagDatabase");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbDeleteObjectTag");
      
      CmMessageInstallHandler (CVSDefaultHandler,  "DbLoadObjByTag");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbLoadObjByDate");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbLoadObjsByTag");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbLoadObjsByDate");
      
      CmMessageInstallHandler (CVSDefaultHandler,  "DbCreateBranch");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbObjectBranch");
      
      CmMessageInstallHandler (CVSDefaultHandler,  "DbCiObjInBranch");

      CmMessageInstallHandler (CVSDefaultHandler,  "DbGetVersionHistory");
      CmMessageInstallHandler (CVSDefaultHandler,  "DbGetObjectHistory");
    }
}  


/*---------------------------------------------------------------------*/
void DbVersioningCleanup (void)
/*---------------------------------------------------------------------*/
{
  if (DbUseCVS ())
    {
      CmMessageInstallHandler (ReleaseHandler, "CVSRelease");
    }
}  


/*---------------------------------------------------------------------*/
void DbSetUseCVS (void)
/*---------------------------------------------------------------------*/
{
  if (UseCVS == 0)
    if (CVSConfigured ())
      UseCVS = 1;
  
  if (DbDebug ())
    {
      DbPrint ("DbSetUseCVS> UseCVS : %d\n", UseCVS);
    }
}


/*---------------------------------------------------------------------*/
void DbSetNoCVS (void)
/*---------------------------------------------------------------------*/
{
  UseCVS = 0;

  if (DbDebug ())
    {
      DbPrint ("DbSetNoCVS> UseCVS : %d\n", UseCVS);
    }
}


/*---------------------------------------------------------------------*/
int DbUseCVS (void)
/*---------------------------------------------------------------------*/
{
  if (DbDebug ())
    {
      DbPrint ("DbUseCVS> UseCVS : %d\n", UseCVS);
    }
  
  return (UseCVS);
}


/*---------------------------------------------------------------------*/
static int CVSConfigured (void)
/*---------------------------------------------------------------------*/
{
  static int  done = 0;
  static char base[] = "%s/%s/%s";
  static char cvs_base[] = "%s/%s/%s";
  static char cvs_name[] = "CVS";
  char*       env;
  char        cvs_repository[256];
  char*       test_file_name = ".dbtest";
  char*       root;
  char*       file_name;
  char*       module_name;
  FILE*       test_file;
  
  if (done) return (1);
  done = 1;

  env = getenv (DBCVSROOT);
  if (env == NULL)
    {
      DbSetNoCVS ();
      return (0);
    }

  strcpy (cvs_repository, env);
  
  root = DbObjectsRoot ();
  
  file_name = (char*) malloc (sizeof(base) +
                              strlen(root) +
                              strlen(cvs_name) +
                              strlen(test_file_name) + 1);
  
  sprintf (file_name, base, root, cvs_name, test_file_name);
  test_file = fopen (file_name, "w");
  if (test_file != NULL)
    {
      fclose (test_file);
      unlink (file_name);
      free (file_name);
    }
  else 
    {
      DbSetErrorInfo (DbErrorCVSRepositoryUndefined,
                      "CVSConfigured> Object repository is not "
                      "known by CVS\n");
    
      free (file_name);
      
      DbSetNoCVS ();
      
      return (0);
    }
  
  module_name = CVSObjectModuleName ();
  
  file_name = (char*) malloc (strlen(cvs_base) +
                              strlen(cvs_repository) +
                              strlen(module_name) + 
                              strlen(test_file_name) + 1);
  
  sprintf (file_name, cvs_base, cvs_repository, module_name, test_file_name);
  test_file = fopen (file_name, "w");
  if (test_file != NULL)
    {
      fclose (test_file);
      unlink (file_name);
      free (file_name);
    }
  else 
    {
      DbSetErrorInfo (DbErrorCVSModuleUnavailable,
                      "CVSConfigured> CVS module %s is not accessible\n",
                      module_name);
      
      free (file_name);
      
      DbSetNoCVS ();
      
      return (0);
    }

  DbSetUseCVS ();
  
  return (1);
}

/*---------------------------------------------------------------------*/
static char* DbObjectsRoot (void)
/*---------------------------------------------------------------------*/
{
  static char temp[200] = "";
  char* root = NULL;
  
  if (strlen (temp) > 0) return (temp);
  
  root = getenv (DBOBJECTSROOT);
  if (root == NULL) root = "./";

  strcpy (temp, root);
  
  return (temp);
}


/*--------------------------------------------------------------------*/
static char* ObjectFileName (char* name)
/*---------------------------------------------------------------------*/
{
  static char  suffix[] = ".db";
  static char* file_name = NULL;
  
  if (file_name == NULL)
    {
      file_name = (char*) malloc (strlen (name) + sizeof (suffix) + 1);
    }
  else
    {
      file_name = (char*) realloc (file_name, 
                                   strlen (name) + sizeof (suffix) + 1);
    }

  strcpy (file_name, name);
  strcat (file_name, suffix);
  
  return (file_name);
}

/*=====================================================================*/

/*---------------------------------------------------------------------*/
static char* CVSObjectFileName (char* name)
/*---------------------------------------------------------------------*/
{
  static char suffix[] = ",v";
  static char* file_name = NULL;
  int size;
  char* object_file_name;
  
  object_file_name = ObjectFileName (name);

  size = strlen (object_file_name) + sizeof (suffix);
  if (file_name == NULL)
    {
      file_name = (char*) malloc (size + 1);
    }
  else
    {
      file_name = (char*) realloc (file_name, size + 1);
    }

  strcpy (file_name, object_file_name);
  strcat (file_name, suffix);
  
  return (file_name);
}

/*---------------------------------------------------------------------*/
static char* CVSRoot (void)
/*---------------------------------------------------------------------*/
{
  static char temp[200] = "";
  char* cvs_root;
  char* default_path;
  
  if (strlen (temp) > 0) return (temp);

  default_path = getenv (CVSROOT);
  cvs_root = getenv (DBCVSROOT);
  if (cvs_root == NULL) cvs_root = default_path;
  if (cvs_root == NULL) cvs_root = "./";

  strcpy (temp, cvs_root);

  return (temp);
}

/*---------------------------------------------------------------------*/
static char* CVSObjectModuleName (void)
/*---------------------------------------------------------------------*/
{
  static char module_name[] = "Objects";
  
  return ((char*) module_name);
}

/*---------------------------------------------------------------------*/
static char* CVSObjectPathAndFileName (char* path, char* name)
/*---------------------------------------------------------------------*/
{
  static char format[] = "%s/%s/%s/%s";
  static char* complete_name = NULL;
  char* cvs_file_name;
  int size;

  cvs_file_name = CVSObjectFileName (name);

  size = sizeof (format) +
      strlen (CVSRoot ()) +
      strlen (CVSObjectModuleName ()) +
      strlen (path) +
      strlen (cvs_file_name);
      
  if (complete_name == NULL)
    {
      complete_name = (char*) malloc (size + 1);
    }
  else
    {
      complete_name = (char*) realloc (complete_name, size + 1);
    }

  sprintf (complete_name, format,
           CVSRoot (), CVSObjectModuleName (), path, cvs_file_name);

  return (complete_name);
}

/*---------------------------------------------------------------------*/
static char* CVSObjectPath (char* path)
/*---------------------------------------------------------------------*/
{
  char* object_path = NULL;
  
  object_path = (char*) malloc (strlen (DbObjectsRoot ()) + 
                                strlen (path) + 1);

  if (object_path != NULL)
    {
      strcpy (object_path, DbObjectsRoot ());
      strcat (object_path, path);
      
      return (object_path);
    }
  else
    return (DbObjectsRoot ());  
}

/*---------------------------------------------------------------------*/
static char* CVSObjectFile (char* path, char* name)
/*---------------------------------------------------------------------*/
{
  char* file_name = NULL;

  file_name = (char*) malloc (strlen (path) + strlen (name) + 1 + 1);
  if (file_name != NULL)
    {
      strcpy (file_name, path);
      if (file_name[strlen (path)] != '/')
        {
          strcat (file_name, "/");
        }
      strcat (file_name, name);
    }
  
  return (file_name);
}

/*=====================================================================*/

/*---------------------------------------------------------------------*/
static void CVSAddPath (char* path)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; $DBROOT/mgr/CVSAddPath.sh %s";

  char* command;
  char* objects_path;
  
  if (!DbUseCVS ()) return;
  if (path == NULL) return;
  
  objects_path = DbObjectsRoot ();
    
  command = (char*) malloc (sizeof (command_line) +
                            strlen (objects_path) +
                            strlen (path) + 1);
  sprintf (command, command_line, objects_path, path);
    
  if (DbDebug ())
    printf ("CVSAddPath> command : %s\n", command);

  system (command);
  
  free (command);
  return;
}

/*---------------------------------------------------------------------*/
static void CVSAdd (char* path, char* name)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s add %s";

  char* command;
  char* complete_file_name;
  char* object_file_name;
  char* object_path;
  char* cvs_root;
  FILE* file;
  
  if (!DbUseCVS ()) return;

  complete_file_name = CVSObjectPathAndFileName (path, name);
  file = fopen (complete_file_name, "r");
  if (file != NULL)
    {
      fclose (file);
      return;
    }

  if ((path != NULL) &&
      (strlen (path) > 0))
    {
      CVSAddPath (path);
    }
  
  cvs_root = CVSRoot ();
  
  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            strlen (cvs_root) +
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, object_file_name);
    
  if (DbDebug ())
    printf ("CVSAdd> %s\n", command);

  system (command);

  free (object_path);
  free (command);

  return;
}

/*---------------------------------------------------------------------*/
void DbCommitObjectVersion (char* path, char* name)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s commit "
      "-m \"%s updated\" %s";

  char* command;
  char* cvs_root;
  char* object_path;
  char* object_file_name;
  
  if (!DbUseCVS ()) return;
  
  CVSAdd (path, name);
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);

  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            strlen (cvs_root) +
                            2 * strlen (object_file_name) + 1);

  sprintf (command, command_line,
           object_path, cvs_root, object_file_name, object_file_name);
  
  if (DbDebug ())
    printf ("DbCommitObjectVersion> %s\n", command);
  
  system (command);
  
  free (object_path);
  free (command);
}

/*---------------------------------------------------------------------*/
static void CVSRelease ()
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s/..; cvs -Q release %s";

  char* command;
  char* cvs_root;
  
  if (!DbUseCVS ()) return;
  
  cvs_root = CVSRoot ();
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (cvs_root) + 1);
  sprintf (command, command_line, cvs_root);
    
  system (command);
  
  if (DbDebug ())
    printf ("CVSRelease> %s\n", command);

  free (command);
  return;
}

/*---------------------------------------------------------------------*/
static DbObject CVSLoadObject (char* command)
/*---------------------------------------------------------------------*/
{
  DbObject object = NULL;
  char* text;

  if (!DbUseCVS ()) return (NULL);
  
  DbClearErrorInfo ();

  if (command == NULL) return (NULL);

  text = DbExecuteCommand (command);
  if (text != NULL)
    {
      object = DbLoadObjectCopyFromText (text);
      free (text);
    }
  else
    {
      DbSetErrorInfo (DbErrorCVSUnloadedObject, "CVSLoadObject> Object not loaded");
    }
  if (DbDebug ())
    printf ("CVSLoadObject> %s\n", command);
  
  return (object);
}


/*---------------------------------------------------------------------*/
static DbObject CVSLoadObjectByDate (char* path, char* name, char* date)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s update -p -D \"%s\" %s";

  DbObject  object = NULL;

  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;
  
  DbClearErrorInfo ();
  
  if (!DbUseCVS ()) return (NULL);

  if ((path == NULL) || (name == NULL) || (date == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSLoadObjectByDate>");
      return (NULL);
    }
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) + 
                            strlen (cvs_root) + 
                            strlen (date) + 
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, date, object_file_name);
  
  object = CVSLoadObject (command);

  free (object_path);
  free (command);

  return (object);
}

/*---------------------------------------------------------------------*/
static DbObject CVSLoadObjectByTag (char* path, char* name, char* tag)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s update -p -r \"%s\" %s";

  DbObject  object = NULL;

  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;
  
  DbClearErrorInfo ();
  
  if (!DbUseCVS ()) return (NULL);
  
  if ((path == NULL) || (name == NULL) || (tag == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSLoadObjectByTag>");
      return (NULL);
    }
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);
  
  command = (char*) malloc (sizeof (command_line) + 
                            strlen (object_path) + 
                            strlen (cvs_root) + 
                            strlen (tag) + 
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, tag, object_file_name);
  
  object = CVSLoadObject (command);
  
  free (object_path);
  free (command);

  return (object);
}


/*---------------------------------------------------------------------*/
static char* CVSLoadObjects (char* path_reg, char* name_reg)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s update -p %s";

  char* text;

  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;
  
  DbClearErrorInfo ();
  
  if (!DbUseCVS ()) return (NULL);

  if ((path_reg == NULL) || (name_reg == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSLoadObjects>");
      return (NULL);
    }
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name_reg);
  object_path      = CVSObjectPath  (path_reg);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) + 
                            strlen (cvs_root) + 
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, object_file_name);
  
  if (DbDebug ())
    printf ("CVSLoadObjectsByDate> %s\n", command);

  text = DbExecuteCommand (command);

  if (text == NULL)
    {
      DbSetErrorInfo (DbErrorCVSUnloadedObject,
                      "CVSLoadObjectsByDate> Object(s) not loaded");
    }
  
  free (object_path);
  free (command);

  if (DbDebug ())
    printf ("CVSLoadObjectsByDate> %s\n", text);

  return (text);
}


/*---------------------------------------------------------------------*/
static char* CVSLoadObjectsByDate (char* path_reg, char* name_reg, char* date)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s update -p -D \"%s\" %s";

  char* text;
  
  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;
  
  DbClearErrorInfo ();
  
  if (!DbUseCVS ()) return (NULL);

  if ((path_reg == NULL) || (name_reg == NULL) || (date == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSLoadObjectsByDate>");
      return (NULL);
    }
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name_reg);
  object_path      = CVSObjectPath  (path_reg);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) + 
                            strlen (cvs_root) + 
                            strlen (date) + 
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, date, object_file_name);
  
  if (DbDebug ())
    printf ("CVSLoadObjectsByDate> %s\n", command);

  text = DbExecuteCommand (command);

  if (text == NULL)
    {
      DbSetErrorInfo (DbErrorCVSUnloadedObject, "CVSLoadObjectsByDate> Object(s) not loaded");
    }
  
  free (object_path);
  free (command);

  if (DbDebug ())
    printf ("CVSLoadObjectsByDate> %s\n", text);

  return (text);
}

/*---------------------------------------------------------------------*/
static char* CVSLoadObjectsByTag (char* path_reg, char* name_reg, char* tag)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s update -p -r \"%s\" %s";

  char* text;
  
  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;
  
  DbClearErrorInfo ();
  
  if (!DbUseCVS ()) return (NULL);
  
  if ((path_reg == NULL) || (name_reg == NULL) || (tag == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSLoadObjectsByTag>");
      return (NULL);
    }
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name_reg);
  object_path      = CVSObjectPath  (path_reg);
  
  command = (char*) malloc (sizeof (command_line) + 
                            strlen (object_path) + 
                            strlen (cvs_root) + 
                            strlen (tag) + 
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, tag, object_file_name);  
  
  if (DbDebug ())
    printf ("CVSLoadObjectsByTag> %s\n", command);

  text = DbExecuteCommand (command);

  if (text == NULL)
    {
      DbSetErrorInfo (DbErrorCVSUnloadedObject,
                      "CVSLoadObjectsByTag> Object(s) not loaded");
    }
  
  free (object_path);
  free (command);

  if (DbDebug ())
    printf ("CVSLoadObjectsByTag> %s\n", text);

  return (text);
}


/*---------------------------------------------------------------------*/
static void CVSTagObject (char* path, char* name, char* tag)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s tag %s %s";

  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;

  FILE*  file;
  
  DbClearErrorInfo ();
  
  if (!DbUseCVS ()) return;

  if ((path == NULL) || (name == NULL)  || (tag == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSTagObject>");
      return;
    }
  
  cvs_root = CVSRoot ();
  
  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);

  file = fopen (CVSObjectFile (object_path, object_file_name), "r");
  if (file == NULL) return;
  fclose (file);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            strlen (cvs_root) +
                            strlen (tag) +
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, tag, object_file_name);
  
  if (DbDebug ())
    printf ("CVSTagObject> %s\n", command);
  
  system (command);

  free (object_path);
  free (command);
  
  return;
}

/*---------------------------------------------------------------------*/
static void CVSTagInPath (char* path, char* tag)
/*---------------------------------------------------------------------*/
/*                                                                     */
/*   On a fait le choix de tagger uniquement le directory local        */
/* (option -l de CVS).                                                 */
/*                                                                     */
/*   Ce choix peut etre modifie                                        */
/* 1) soit on autorise la recursivite                                  */
/* 2) soit on installe un mode configurable                            */
/*                                                                     */
/* (qui pourrait eventuellement supprimer la necessite de TagAll)      */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s tag -l %s";

  char* command;
  char* object_path;
  char* cvs_root;

  if (!DbUseCVS ()) return;

  if ((path == NULL) || (tag == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSTagInPath>");
      return;
    }
  
  cvs_root    = CVSRoot       ();
  object_path = CVSObjectPath (path);

  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            strlen (cvs_root) +
                            strlen (tag) + 1);
  sprintf (command, command_line, object_path, cvs_root, tag);
  
  if (DbDebug ())
    printf ("CVSTagInPath> %s\n", command);
  
  system (command);
  
  free (object_path);
  free (command);

  return;
}

/*---------------------------------------------------------------------*/
static void CVSTagAll (char* tag)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s tag -R %s";

  char* command;
  char* object_path;
  char* cvs_root;

  if (!DbUseCVS ()) return;

  if (tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSTagAll>");
      return;
    }
  
  cvs_root    = CVSRoot        ();
  object_path = DbObjectsRoot ();

  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            strlen (cvs_root) +
                            strlen (tag) + 1);
  sprintf (command, command_line, object_path, cvs_root, tag);
  
  if (DbDebug ())
    printf ("CVSTagAll> %s\n", command);
  
  system (command);
  
  free (command);

  return;
}

/*---------------------------------------------------------------------*/
static void CVSDeleteTag (char* path, char* name, char* tag)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s tag -d %s %s";

  char* command;
  char* object_file_name;
  char* object_path;
  char* cvs_root;

  FILE*  file;
  
  if (!DbUseCVS ()) return;

  if ((path == NULL) || (name == NULL)  || (tag == NULL))
    {
      DbSetErrorInfo (DbErrorBadArgument, "CVSDeleteTag>");
      return;
    }
  
  cvs_root = CVSRoot ();
  
  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);

  file = fopen (CVSObjectFile (object_path, object_file_name), "r");
  if (file == NULL) return;
  fclose (file);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            strlen (cvs_root) +
                            strlen (tag) +
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, object_path, cvs_root, tag, object_file_name);
  
  if (DbDebug ())
    printf ("CVSDeleteTag> %s\n", command);
  
  system (command);

  free (object_path);
  free (command);
  
  return;
}

/*---------------------------------------------------------------------*/
static void CVSRemovePath (char* path)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; $DBROOT/mgr/CVSRemovePath.sh %s";

  char* command;
  char* objects_path;
  
  if (!DbUseCVS ()) return;
  if (path == NULL) return;
  
  objects_path = DbObjectsRoot ();
    
  command = (char*) malloc (sizeof (command_line) +
                            strlen (objects_path) +
                            strlen (path) + 1);
  sprintf (command, command_line, objects_path, path);
    
  if (DbDebug ())
    printf ("CVSRemovePath> command : %s\n", command);

  system (command);
  
  free (command);
  return;
}

/*---------------------------------------------------------------------*/
void DbRemoveObjectVersion (char* path, char* name)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s remove %s; "
      "cvs -Q -d %s commit -m \"%s removed\" %s";

  char* command;
  char* cvs_root;
  char* object_path;
  char* object_file_name;
  
  if (!DbUseCVS ()) return;
  
  cvs_root = CVSRoot ();

  object_file_name = ObjectFileName (name);
  object_path = CVSObjectPath (path);

  command = (char*) malloc (sizeof (command_line) +
                            strlen (object_path) +
                            2*strlen (cvs_root) +
                            3*strlen (object_file_name) + 1);
  sprintf (command, command_line,
           object_path, cvs_root, object_file_name,
           cvs_root, object_file_name, object_file_name);
  
  if (DbDebug ())
    printf ("CVSRemoveObject> %s\n", command);
  
  system (command);

  if ((path != NULL) &&
      (strlen (path) > 0))
    {
      CVSRemovePath (object_path);
    }
  
  free (object_path);
  free (command);
}

/*---------------------------------------------------------------------*/
static void CVSCreateBranch (char* tag, char* branch)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s rtag -r %s -b %s %s";

  char* command;
  char* objects_root;
  char* cvs_root;
  char* module_name;
  FILE* file;
  
  if (!DbUseCVS ()) return;
  
  objects_root = DbObjectsRoot  ();
  cvs_root     = CVSRoot ();
  module_name  = CVSObjectModuleName  ();
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (objects_root) +
                            strlen (cvs_root) +
                            strlen (tag) +
                            strlen (branch) +
                            strlen (module_name) + 1);
  sprintf (command, command_line, objects_root, cvs_root,
           tag, branch, module_name);
  
  if (DbDebug ())
    printf ("CVSCreateBranch> %s\n", command);
  
  system (command);
  
  free (command);
  return;
}

/*---------------------------------------------------------------------*/
static void CVSObjectBranch (char* name, char* branch)
/*---------------------------------------------------------------------*/
{
  static char command_line[] = "cd %s; cvs -Q -d %s tag -b -r %s %s";

  char* command;
  char* objects_root;
  char* cvs_root;
  char* complete_file_name;
  char* object_file_name;
  FILE* file;
  
  if (!DbUseCVS ()) return;
  
  objects_root = DbObjectsRoot ();
  cvs_root     = CVSRoot ();
  object_file_name = ObjectFileName (name);
  
  complete_file_name = CVSObjectPathAndFileName (objects_root, name);
  file = fopen (complete_file_name, "r");
  if (file == NULL) return;
  fclose (file);
  
  command = (char*) malloc (sizeof (command_line) +
                            strlen (objects_root) +
                            strlen (cvs_root) +
                            strlen (branch) +
                            strlen (object_file_name) + 1);
  sprintf (command, command_line, objects_root, cvs_root, branch, 
           object_file_name);
  
  if (DbDebug ())
    printf ("CVSObjectBranch> %s\n", command);
  
  system (command);
  
  free (command);
  return;
}


/*---------------------------------------------------------------------*/
static void CVSCommitObjectInBranch (DbObject object, char* branch)
/*---------------------------------------------------------------------*/
{
#define Db_tmp .db_tmp_for_branch
  
  static char command_line[] = "${DBROOT}/mgr/CVSCommitObjectInBranch.sh '%s' %s %s";

  if (!DbUseCVS ()) return;

  DbClearErrorInfo ();
    
  {
    char* cvs_root;
    char* objects_root;
    char* object_path;
    char* cvs_object_path;
    char* object_file_name;

    char* file_name;
    
    FILE* file;
    char* command;

    cvs_root = CVSRoot ();
    objects_root  = DbObjectsRoot ();
    
    object_file_name = ObjectFileName (DbObjectGetName (object));
    object_path      = DbObjectGetPathName (object);
    cvs_object_path  = CVSObjectPath (object_path);
      
    if (DbDebug ())
      {
        DbPrint ("CVSCommitObjectInBranch> Object : %s/%s - Branch : %s\n",
                 cvs_object_path, object_file_name, branch);
      }
    
    file_name = BuildFileName (DbObjectGetName(object), cvs_object_path);
    
    file = fopen (CVSObjectFile (cvs_object_path, object_file_name), "r");
    if (file == NULL) return;
    fclose (file);

/* Update object */
    {
      char* text;
      char* tmp_file_name;

      tmp_file_name = BuildTmpFileName (DbObjectGetName(object), 
                                        cvs_object_path,
                                        branch);

      text = DbConvertObjectToString (object);
      if (DbDebug ())
        {
          DbPrint ("CVSCommitObjectInBranch> text : %s\n", text);
        }

      DbWriteFile (tmp_file_name, text);
    }
    
/* Load object in branch */
    command = (char*) malloc (sizeof (command_line) +
                              strlen (object_path) +
                              strlen (object_file_name) +
                              strlen (branch) + 1);
    sprintf (command, command_line, object_path, object_file_name, branch);
    
    if (DbDebug ())
      printf ("CVSCommitObjectInBranch>command : %s\n", command);
    
    system (command);
    free (command);
  }
  
  return;
}


/*= CVS handlers ======================================================*/


/*---------------------------------------------------------------------*/
static CmMessageStatus SetVersioningHandler (CmMessage message,
                                             char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;
  
  if (DbDebug ()) DbPrint ("SetVersioningHandler> from %s\n", sender);
  
  DbClearErrorInfo ();
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  DbSetUseCVS ();
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus SetNoVersioningHandler (CmMessage message,
                                               char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;
  
  if (DbDebug ()) DbPrint ("SetNoVersioningHandler> from %s\n", sender);
  
  DbClearErrorInfo ();
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbUseCVS ())
    {
      DbSetNoCVS ();
    }
    
  return (CmMessageOk);
}


/*---------------------------------------------------------------------*/
static CmMessageStatus GetVersioningHandler (CmMessage message,
                                             char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  CmMessage answer;
  int       useCVS;
  int transaction = 0;
  
  if (DbDebug ()) DbPrint ("GetVersioningHandler> from %s\n", sender);
  
  DbClearErrorInfo ();

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  useCVS = DbUseCVS ();

  if (DbDebug ())
    {
      printf ("GetVersioningHandler>  useCVS : %d\n", useCVS);
    }
  
  answer = CmMessageNew ();
  CmMessagePutInt  (answer, useCVS);
  CmMessageSetType (answer, "DbVersioning");
  CmMessageSend    (answer, sender);
  CmMessageDelete (answer);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus ReleaseHandler (CmMessage message,
                                       char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  int transaction = 0;

  if (DbDebug ())
    {
      printf ("ReleaseHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  CVSRelease ();
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}


/*---------------------------------------------------------------------*/
static CmMessageStatus CVSDefaultHandler (CmMessage message,
                                          char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  int transaction = 0;

  if (DbDebug ())
    {
      printf ("CVSDefaultHandler> from %s\n", sender);
    }
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  DbSetErrorInfo (DbErrorCVSNotAvailable, "CVSDefaultHandler>");
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*= Tag Handlers ======================================================*/

/*---------------------------------------------------------------------*/
static CmMessageStatus TagObjectHandler (CmMessage message,
                                         char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* name;
  char* path;
  char* tag;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("TagObjectHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  name = CmMessageGetText (message);
  path = CmMessageGetText (message);
  tag  = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  CVSTagObject (path, name, tag);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus TagObjectsInPathHandler (CmMessage message,
                                                char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* path;
  char* object_tag;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("TagObjectsInPathHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  path       = CmMessageGetText (message);
  object_tag = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  CVSTagInPath (path, object_tag);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus TagDatabaseHandler (CmMessage message,
                                             char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* tag;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("TagDatabaseHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  tag = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  CVSTagAll (tag);  
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus DeleteObjectTagHandler (CmMessage message,
                                               char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* name;
  char* path;
  char* tag;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("DeleteObjectTagHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  name = CmMessageGetText (message);
  path = CmMessageGetText (message);
  tag  = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  CVSDeleteTag (path, name, tag);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*= Load Handlers =====================================================*/

/*---------------------------------------------------------------------*/
static CmMessageStatus LoadObjectByTagHandler (CmMessage message,
                                               char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  DbObject object = NULL;
  char* object_name;
  char* object_path;
  char* object_tag;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("LoadObjectByTagHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  object_name = CmMessageGetText (message);
  object_path = CmMessageGetText (message);
  object_tag  = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  GCReferenceAssign (object, CVSLoadObjectByTag (object_path, 
                                                 object_name, object_tag));
  
  if (object != NULL)
    {
      CmMessage answer;
      char* text;
      
      answer = CmMessageNew ();
      text   = DbConvertObjectToString (object);
      CmMessagePutText (answer, text);
      CmMessageSetType (answer, "DbObjectDefinition");
      CmMessageSend    (answer, sender);
      CmMessageDelete (answer);
    }
  else
    {
      DbSetErrorInfo (DbGetError (), "LoadObjectByTagHandler>");
    }

  DbSendBreak (sender, transaction);
  
  GCReferenceClear (object);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus LoadObjectByDateHandler (CmMessage message,
                                                char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  DbObject object = NULL;
  char* object_name;
  char* object_path;
  char* object_date;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("LoadObjectByDateHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  object_name = CmMessageGetText (message);
  object_path = CmMessageGetText (message);
  object_date = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  GCReferenceAssign (object, CVSLoadObjectByDate (object_path, 
                                                  object_name, object_date));
  
  if (object != NULL)
    {
      CmMessage answer;
      char*     text;
      
      answer = CmMessageNew ();
      text = DbConvertObjectToString (object);
      CmMessagePutText (answer, text);
      CmMessageSetType (answer, "DbObjectDefinition");
      CmMessageSend (answer, sender);
      CmMessageDelete (answer);
    }
  else
    {
      DbSetErrorInfo (DbGetError (), "LoadObjectByDateHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  GCReferenceClear (object);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus LoadObjectFromBranchHandler (CmMessage message,
                                                    char* sender, 
                                                    char* server)
/*---------------------------------------------------------------------*/
{
  DbObject object = NULL;
  char* object_name;
  char* object_path;
  char* object_branch;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("LoadObjectFromBranchHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  object_name   = CmMessageGetText (message);
  object_path   = CmMessageGetText (message);
  object_branch = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if ((object_name != NULL) && (strlen (object_name) > 0))
    {
      char* full_name = NULL;
      DbBranch branch;
      DbLock lock;
      DbClient client;
      
      DbClearErrorInfo ();

      full_name = BuildFullName (object_name, object_path);

      lock = DbFindLockInBranch (object_branch, full_name);
      client = DbFindClient (sender);

        /*
          Objects are always at least readable but we have to
          signal when a ReadWrite lock is set AND the requested
          object is already locked by another client.
          */

      if (DbClientRequestWriteAccess (client) &&
          !DbLockPermitsReadForWrite (lock, sender))
        {
          if (lock == NULL)
            {
              DbSetErrorInfo (DbErrorWriteAccessDenied, 
                              "LoadObjectFromBranchHandler> not locked");
            }
          else
            {
              DbSetErrorInfo (DbErrorWriteAccessDenied, 
                              "LoadObjectFromBranchHandler> locked by %s", 
                              DbLockGetOwner (lock));
            }
        }
      else
        {
          GCReferenceAssign (object, 
                             CVSLoadObjectByTag (object_path, 
                                                 object_name, 
                                                 object_branch));
          
          if (object == NULL)
            {
              DbSetErrorInfo (DbGetError (), 
                              "LoadObjectFromBranchHandler> cvs error");
            }
        }
      
      if (object != NULL)
        {
          CmMessage answer;
          char* text;
          
          if (DbClientRequestWriteAccess (client))
            {
                /*
                  When the sender is declared as a client, AND that
                  it has set ReadWrite locks, AND that this point has
                  been reached, this implies that a lock must be
                  set on this object.
                  */
              DbLockNewInBranch (object_branch, full_name, client);
            }
          
          answer = CmMessageNew ();
          text   = DbConvertObjectToString (object);
          CmMessagePutText (answer, text);
          CmMessageSetType (answer, "DbObjectDefinition");
          CmMessageSend    (answer, sender);
          CmMessageDelete (answer);
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "LoadObjectFromBranchHandler>");
    }

  DbSendBreak (sender, transaction);
  
  GCReferenceClear (object);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus LoadObjectsByTagHandler (CmMessage message,
                                                char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  
  char* path_reg;
  char* name_reg;
  char* tag;
  int   transaction = 0;
  
  if (DbDebug ())
    {
      printf ("LoadObjectsByTagHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  name_reg = CmMessageGetText (message);
  path_reg = CmMessageGetText (message);
  tag      = CmMessageGetText (message);
  
  if (DbDebug ())
  {
    printf ("LoadObjectsByTagHandler> path : %s - objects : %s - tag : %s\n",
            path_reg, name_reg, tag);
  }
      
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  text = CVSLoadObjectsByTag (path_reg, name_reg, tag);

  if (text != NULL)
  { /* Construction et envoi du message reponse */
    CmMessage answer;
    CLList    objects = NULL;
    CIterator it;

    GCReferenceAssign (objects, CLListNew ());

    DbLoadObjectCopiesFromText (text, objects);

    answer = CmMessageNew ();
    CmMessagePutInt  (answer, transaction);

    it = CLListGiveIterator (objects);

    for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbObject object = NULL;

      object = (DbObject) CIteratorItem (it);
      
      if (object != NULL)
        {
          char* temp;

          temp = DbConvertObjectToString (object);
          
          if (DbDebug ())
          {
            printf ("LoadObjectsByTagHandler> obj_txt : [%s]\n", temp);
          }
          
          CmMessagePutText (answer, temp);
        }
    }
    CIteratorDelete  (it);
    CLListErase      (objects);
    GCReferenceClear (objects);
    
    CmMessageSetType (answer, "DbObjectDefs");
    CmMessageSend    (answer, sender);
    CmMessageDelete  (answer);

    free (text);
  }
  else
  {
    DbSetErrorInfo (DbGetError (), "LoadObjectsByTagHandler>");
  }

  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus LoadObjectsByDateHandler (CmMessage message,
                                                 char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  
  char* path_reg;
  char* name_reg;
  char* date;
  int   transaction = 0;
  
  if (DbDebug ())
    {
      printf ("LoadObjectsByDateHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  name_reg = CmMessageGetText (message);
  path_reg = CmMessageGetText (message);
  date     = CmMessageGetText (message);
  
  if (DbDebug ())
  {
    printf ("LoadObjectsByDateHandler> path : %s - objects : %s - date : %s\n",
            path_reg, name_reg, date);
  }
      
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  text = CVSLoadObjectsByDate (path_reg, name_reg, date);

  if (text != NULL)
  { /* Construction et envoi du message reponse */
    CmMessage answer;
    CLList    objects = NULL;
    CIterator it;

    GCReferenceAssign (objects, CLListNew ());

    DbLoadObjectCopiesFromText (text, objects);

    answer = CmMessageNew ();
    CmMessagePutInt  (answer, transaction);

    it = CLListGiveIterator (objects);

    for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbObject object = NULL;

      object = (DbObject) CIteratorItem (it);
      
      if (object != NULL)
        {
          char* temp;

          temp = DbConvertObjectToString (object);
          
          if (DbDebug ())
          {
            printf ("LoadObjectsByDateHandler> obj_txt : [%s]\n", temp);
          }
          
          CmMessagePutText (answer, temp);
        }
    }
    CIteratorDelete  (it);
    CLListErase      (objects);
    GCReferenceClear (objects);
    
    CmMessageSetType (answer, "DbObjectDefs");
    CmMessageSend    (answer, sender);
    CmMessageDelete  (answer);

    free (text);
  }
  else
  {
    DbSetErrorInfo (DbGetError (), "LoadObjectsByDateHandler>");
  }

  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*=====================================================================*/

/*---------------------------------------------------------------------*/
static CmMessageStatus CreateBranchHandler (CmMessage message,
                                            char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* tag;
  char* branch;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("CreateBranchHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  tag     = CmMessageGetText (message);
  branch  = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if ((tag != NULL) && (branch != NULL))
    CVSCreateBranch (tag, branch);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus ObjectBranchHandler (CmMessage message,
                                            char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* object_name;
  char* branch;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("ObjectBranchHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  object_name    = CmMessageGetText (message);
  branch         = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if ((object_name != NULL) && (branch != NULL))
    CVSObjectBranch (object_name, branch);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus CommitObjectInBranchHandler (CmMessage message,
                                                    char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  char* branch;
  int transaction = 0;
  
  if (DbDebug ())
    {
      printf ("CommitObjectInBranchHandler> from %s\n", sender);
    }

  text   = CmMessageGetText (message);
  branch = CmMessageGetText (message);
    
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if ((text != NULL) && (branch != NULL))
    {
      DbObject object;
      
      DbClearErrorInfo ();
      
      if (DbDebug ())
        {
          printf ("CommitObjectInBranchHandler> text : %s\n", text);
        }
      
      object = DbParserCreateObject (text);
      
      if (object == NULL)
        {
            /*
              Error while defining the object.
              (syntax error or bad redefinition of an existing object)
              */
          DbSetErrorInfo (DbErrorCVSUnloadedObject,
                          "CommitObjectInBranchHandler>");
        }
      else
        {
          CVSCommitObjectInBranch (object, branch);      
        }
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*=====================================================================*/

/*--------------------------------------------------------------------*/
static CmMessageStatus GetVersionHistoryHandler (CmMessage message, 
                                                 char* sender, char* server)
/*--------------------------------------------------------------------*/
{
  char*     name;
  CmMessage answer = NULL;
  char      command[256];
  char      line[256];
  FILE*     history;
  int transaction = 0;
  
  name = CmMessageGetText (message);

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("GetVersionHistoryHandler> from %s for %s\n", 
              sender, name);
    }
  
  DbClearErrorInfo ();
        
  if (DbUseCVS ())
    {
      if (name != NULL)
        {
          answer = CmMessageNew ();
          
          sprintf (command, "$DBROOT/mgr/CVSShowHistory.sh %s.db", name);
          
          history = popen (command, "r");
          if (history != NULL)
            {
              while (!feof (history))
                {
                  char* ptr;
                  char* nl;
                  
                  line[0] = 0;
                  
                  ptr = fgets (line, sizeof (line), history);
                  if (ptr == NULL) break;
                  nl = strchr (line, '\n');
                  if (nl != NULL) *nl = 0;
                  
                  CmMessagePutText (answer, line);
                  if (DbDebug ())
                    {
                      printf ("GetVersionHistoryHandler> line : %s\n", line);
                    }
                }    
              pclose (history);
            }
          
          CmMessageSetType (answer, "DbVersionHistory");
          CmMessageSend    (answer, sender);
          CmMessageDelete  (answer);
        }
      else
        {
          DbSetErrorInfo (DbErrorBadArgument, "GetVersionHistoryHandler>");
        }
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}


/*--------------------------------------------------------------------*/
static CmMessageStatus GetObjectHistoryHandler (CmMessage message, 
                                                char* sender, char* server)
/*--------------------------------------------------------------------*/
{
  char*     name;
  CmMessage answer = NULL;
  char      command[256];
  char      line[256];
  FILE*     history;
  int transaction = 0;
  
  name = CmMessageGetText (message);

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("GetObjectHistoryHandler> from %s for %s\n", 
              sender, name);
    }
  
  if (name != NULL)
    {
      answer = CmMessageNew ();
      
      DbClearErrorInfo ();
        /*
          Rcupration des versions
          */
    
      CmMessageSetType (answer, "DbObjectHistory");
      CmMessageSend    (answer, sender);
      CmMessageDelete  (answer);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "GetObjectHistoryHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static char* DbExecuteCommand (char* command)
/*---------------------------------------------------------------------*/
{
  char line[256];
  FILE* file;
  char* text = NULL;
  
  file = popen (command, "r");
  
  if (file == NULL) return (NULL);

  while (!feof (file))
    {
      char* ptr;
      char* nl;
      int   length;
      
      line[0] = 0;
      
      ptr = fgets (line, sizeof (line), file);
      if (ptr == NULL) break;
      
      length = strlen (line);
      
      if (text == NULL)
        {
          text = (char*) malloc (length * sizeof (char) + 1);
          if (!strcpy (text, line)) return (NULL);
        }
      else
        {
          length += strlen (text);
          text = (char*) realloc (text, length * sizeof (char) + 1);
          if (!strcat (text, line)) return (NULL);
        }
    }
  
  pclose (file);
  
  return (text);
}

