/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <DbObject.h>

/*---------------------------------------------------------------------------*/
static int MatrixCount         = 0;
static int MatrixContentsCount = 0;
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
DbMatrix* DbMatrixNew (DbType element, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  DbMatrix* object = NULL;
  int element_size;

  if (element == NULL) return (NULL);
  if ((rows == 0) || (columns == 0)) return (NULL);

  element_size = DbTypeSize (element);
  if (element_size == 0) return (NULL);

  object = (DbMatrix*) malloc (sizeof (DbMatrix));
  MatrixCount++;

  DbMatrixInit (object, element, rows, columns);

  return (object);
}

/*---------------------------------------------------------------------------*/
void DbMatrixDelete (DbMatrix* object)
/*---------------------------------------------------------------------------*/
{
  DbMatrixErase (object);

  free (object);
  MatrixCount--;
}

/*---------------------------------------------------------------------------*/
void DbMatrixInit (DbMatrix* object, DbType element, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  int element_size;
  int number;
  char* address;

  if (object == NULL) return;

  object->element_handle = NULL;
  object->rows           = rows;
  object->columns        = columns;
  object->contents       = NULL;

  element_size = DbTypeSize (element);
  if (element_size == 0) return;

  object->element_handle = DbTypeHandleNew ();
  DbHandleSet (object->element_handle, element);

  object->contents = (void*) malloc (rows * columns * element_size);
  MatrixContentsCount++;

  memset (object->contents, 0, rows * columns * element_size);

  address = (char*) object->contents;

  for (number = 0; number < (rows * columns); number++)
    {
      DbTypeInit (element, address);
      address += element_size;
    }
}

/*---------------------------------------------------------------------------*/
void DbMatrixErase (DbMatrix* object)
/*---------------------------------------------------------------------------*/
{
  DbType element;

  element = DbMatrixGetElementType (object);
  if (element != NULL)
    {
      int n;
      char* address = (char*) object->contents;
      int element_size = DbTypeSize (element);

      for (n = 0; n < (object->rows * object->columns); n++)
        {
          DbTypeErase (element, address);
          address += element_size;
        }
    }

  if (object->contents != NULL)
    {
      free (object->contents);
      MatrixContentsCount--;
      object->contents = NULL;
    }

  object->rows = 0;
  object->columns = 0;

  DbTypeHandleRemove (object->element_handle);
  object->element_handle = NULL;
}

/*---------------------------------------------------------------------------*/
DbType DbMatrixGetElementType (DbMatrix* object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);
  if (object->element_handle == NULL) return (NULL);

  return ((DbType) DbHandleGet (object->element_handle));
}

/*---------------------------------------------------------------------------*/
int DbMatrixGetRows (DbMatrix* object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (0);

  return (object->rows);
}

/*---------------------------------------------------------------------------*/
int DbMatrixGetColumns (DbMatrix* object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (0);

  return (object->columns);
}

/*---------------------------------------------------------------------------*/
void* DbMatrixGetAddress (DbMatrix* object, int row, int column)
/*---------------------------------------------------------------------------*/
{
  char* result;
  int index;
  DbType element;

  element = DbMatrixGetElementType (object);
  if (element == NULL) return (NULL);

  if (object->contents == NULL) return (NULL);
  if ((row < 0) ||
      (row >= object->rows)) return (NULL);
  if ((column < 0) ||
      (column >= object->columns)) return (NULL);

  index = (row * object->columns) + column;

  result = (char*) object->contents;

  return (result + (index * DbTypeSize (element)));
}

