#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <Db.h>

/*---------------------------------------------------------------------------*/
/*                                                                           */
/*  Visitor to store in file        	                                     */
/*                                                                           */
/*---------------------------------------------------------------------------*/

typedef struct _FilerRec* Filer;
typedef struct _FilerRec
{
  DbVisitorRec base;
  int tab;
  FILE* file;
} FilerRec;

/*---------------------------------------------------------------------------*/
static void tab (Filer visitor)
/*---------------------------------------------------------------------------*/
{
  int i;

  for (i = 0; i < visitor->tab; i++)
    {
      fprintf (visitor->file, "  ");
    }
}

/*---------------------------------------------------------------------------*/
static int down (Filer visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  if (visitor == NULL) return (0);

  tab (visitor);
  fprintf (visitor->file, "{\n");
  visitor->tab++;

  return (1);
}

/*---------------------------------------------------------------------------*/
static void clear (Filer visitor)
/*---------------------------------------------------------------------------*/
{
  if (visitor == NULL) return;

  if (visitor->file != NULL)
    {
      fclose (visitor->file);
      visitor->file = NULL;
    }
}

/*---------------------------------------------------------------------------*/
static int in (Filer visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  void* address;

  if (visitor == NULL) return (0);

  address = DbTypeIteratorGetItemAddress (it);
  if (address != NULL)
    {
      DbType type;
      char* text;
      char* key;
      
      type    = DbTypeIteratorGetItemType (it);
      
      text = DbTypeOutput (type, address);

      if (text != NULL)
        {
          tab (visitor);

          key = DbTypeIteratorGetKey (it);
          if (key != NULL)
            {
              fprintf (visitor->file, key);
              fprintf (visitor->file, " = ");
            }
          
          fprintf (visitor->file, "%s\n", text);
        }
    }

  return (1);
}

/*---------------------------------------------------------------------------*/
static int up (Filer visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  if (visitor == NULL) return (0);

  visitor->tab--;
  tab (visitor);
  fprintf (visitor->file, "}\n");

  return (1);
}

/*---------------------------------------------------------------------------*/
static DbVisitor FilerNew (DbObject object)
/*---------------------------------------------------------------------------*/
{
  Filer filer = NULL;
  char* type_name;
  char* name;
  char* file_name;
  char* suffix = ".db";
  FILE* file;

  if (object == NULL) return (NULL);

  type_name = DbTypeName (DbObjectGetType (object));
  name = DbObjectGetName (object);

  file_name = (char*) malloc (strlen (name) + strlen (suffix) + 1);
  strcpy (file_name, name);
  strcat (file_name, suffix);

  if ((file = fopen (file_name, "w")) == NULL) 
    {
      fprintf (stderr, "Impossible to create file %s\n", file_name);
    } 
  else
    {
      filer = (Filer) malloc (sizeof (FilerRec));

      DbVisitorInit ((DbVisitor) filer,
                     (DbVisitorFunction) clear, 
                     (DbVisitorFunction) down, 
                     (DbVisitorFunction) in, 
                     (DbVisitorFunction) up);

      filer->tab = 0;
      filer->file = file;

      fprintf (filer->file, "%s %s = ", type_name, name);
    }

  free (file_name);

  return ((DbVisitor) filer);
}

/*---------------------------------------------------------------*/
int DbCreateFileFromObject (DbObject object)
/*---------------------------------------------------------------*/
{
  DbVisitor visitor;

  if (object == NULL) return (0);

  visitor = FilerNew (object);
  DbObjectVisit (object, visitor);
  DbVisitorDelete (visitor);

  return (1);
}

